/*
 * Written by Zhen Liu, University of Southern California.
 * $Id: tcp_vegas.h,v 1.3 1995/01/29 00:25:58 zhenliu Exp zhenliu $
 */

#ifndef _netinet_tcp_vegas_h
#define _netinet_tcp_vegas_h

/*
 * TCP connection statistics, 11/26/93
 */
struct connstat {
	u_long	tcp_segstimed;		/* segs where we tried to get rtt */
	u_long	tcp_rttupdated;	        /* times we succeeded */
	u_long	tcp_rexmttimeo;	        /* retransmit timeouts */

	u_long	tcp_sndtotal;		/* total packets sent */
	u_long	tcp_sndpack;		/* data packets sent */
	u_long	tcp_sndbyte;		/* data bytes sent */
	u_long	tcp_sndrexmitpack;	/* data packets retransmitted */
	u_long	tcp_sndrexmitbyte;	/* data bytes retransmitted */
	u_long	tcp_sndacks;		/* ack-only packets sent */
	u_long	tcp_sndprobe;		/* window probes sent */
	u_long	tcp_sndurg;		/* packets sent with URG only */
	u_long	tcp_sndwinup;		/* window update-only packets sent */
	u_long	tcp_sndctrl;		/* control (SYN|FIN|RST) packets sent */
	u_long	tcp_rcvtotal;		/* total packets received */
	u_long	tcp_rcvpack;		/* packets received in sequence */
	u_long	tcp_rcvbyte;		/* bytes received in sequence */
	u_long	tcp_rcvduppack;	        /* duplicate-only packets received */
	u_long	tcp_rcvdupbyte;	        /* duplicate-only bytes received */
	u_long	tcp_rcvpartduppack;	/* packets with some duplicate data */
	u_long	tcp_rcvpartdupbyte;	/* dup. bytes in part-dup. packets */
	u_long	tcp_rcvdupack;		/* rcvd duplicate acks */
	u_long	tcp_rcvackpack;	        /* rcvd ack packets */
	u_long	tcp_rcvackbyte;	        /* bytes acked by rcvd acks */

/* for Vegas-0.8 only, 1/21/95 */
	u_long  tcp_dupftimeo;		/* Fine grained timeout on dup ACK */
	u_long  tcp_worryftimeo;	/* Fine grained timeout on worry */
};


/*
 * Tcp vegas control block, 1/21/95. This is based on Vegas-0.8, 
 * received at 11/94.
 */
struct vcb {
  	char 	v_flags;		/* TCP_TAHOE, TCP_RENO, TCP_VEGAS */
	int	v_rexmtthresh;		/* dupacks recvd before retransmit,
					   default is 3 */
	int	v_slowstart;		/* packets to send after slowstart,
					   default is 1 */

        /* Vegas timer stuff */         
        int     v_rtt;                     
        int     v_sa;
        int     v_sd;                      
        int     v_timeout;              
        int     v_worried;		/* While > 0, check for packet 
					 * timeout */
        int     v_time_cwin_chg;	/* Time when congestion window
					 * was last decreased. Used to prevent
					 * decreasing cwin more than once
					 * for losses which occurred at
					 * same cwin level */
        int     v_synsent;		/* Time when SYN was sent. Used to
					 * set v_rtt at beg of connection */
	int	v_newcwnd;		/* Stores new uninflated cwnd, used
					 * to reset cwnd later. Same function
					 * as snd_ssthresh, but now 
					 * snd_ssthresh is only used during
					 * slow-start */

        /* Vegas exp decrease in rate. Option to decrease window exponentially
	 * (sends on every other ACK) rather than waiting 1/2 RTT to start
	 * sending. NOT USED anymore (better to be conservative) 
	 */
        int     v_send_toggle;          
        int     v_use_sndwnd;           
        int     v_max_send;             

	/* New Vegas stuff (12/14/93) */
	/* Protect from spikes. Limits output rate to at most twice the
	 * current rate. 
	 * It helps for fairness, and results in a little less losses,
	 * but there is a problem when the receiver doesn't ACK every
	 * packet immediately (delayed ACKs) resulting in the sender
	 * not being able to keep up with the congestion window. 
	 * Therefore it is OFF by default, and my be removed in the final
	 * version.
	 */
	int	v_spike_do;
	int	v_spike_nextTime;
	int	v_spike_timeInc;
	int	v_spike_byteCnt;

	/* Exp increase every other during slow-start. */
	int	v_exp_inc_do;
	int	v_exp_inc_flag;
	int	v_exp_inc_cnt;
	int	v_exp_inc_isless;
	int	v_exp_inc_nseg;

	/* Congestion detection (or more correctly avoidance ) variables.
	 * The mechanism checks once per RTT. Current RTT began with sequence
	 * v_cong_detect_begseq 
	 */
	int	v_cong_detect_do;	      	/* Whether to do it */
	tcp_seq	v_cong_detect_begseq;		
	int	v_cong_detect_begtime;		/* When RTT began */
	int	v_cong_detect_predict_do;	/* Whether to predict the
						 * available bw during 
						 * slow-start */
	int	v_cong_detect_last_sendtime;
	int	v_cong_detect_baseRtt;		/* Current base RTT, usually
						 * the minimum */
	int	v_cong_detect_sumRTT;		/* Used to get average */
	int	v_cong_detect_cntRTT;		/* RTT during a RTT interval */

	/* The congestion window should stay within band. Going above
	 * it means that not enough buffers are being used at the
	 * bottleneck, so cwnd should increase. Going below the band means 
	 * that too many buffers are being used, so cwnd should decrease. 
	 */
	int	v_cong_detect_top_isless;	/* => cwnd < top of band */
	int	v_cong_detect_top_nseg;
	int	v_cong_detect_bot_isless;	/* => cwnd < bot of band */
	int	v_cong_detect_bot_nseg;
	int	v_incr;				/* How much to increase cwnd
						 * when an ACK is received
						 */

/* vegas timestamp queue*/
	struct timeval v_start;		/* starting time of the vegas session */
	struct vstamp *v_queue;		/* vegas queue is a list of vegas timestamps
					   of current outstanding segments */
#define POOLSIZE 4096			
	struct vstamp *q_array;		/* vegas queue is implemented as an
					   array of vstamps */
	struct vstamp *q_free;		/* list of un-used vstamps */
	int	q_maxlen;		/* max len of the queue */
};


/* 
 * Some handy macros.
 */
#define VP 		(tp->t_vcb)
#define ISTAHOE(tp)	((tp)->t_vcb->v_flags==TCP_TAHOE)
#define ISRENO(tp)	((tp)->t_vcb->v_flags==TCP_RENO)
#define ISVEGAS(tp) 	((tp)->t_vcb->v_flags==TCP_VEGAS)


/* 
 * Vegas stamps.
 */
struct vstamp {
   	int vs_len;			/* segment length */
	int vs_sendtime;		/* the time when segment was
					   sent. milliseconds from the
					   start of the converstaion */
	int vs_transmits;		/* how many times this segment
					   is sent */
	struct vstamp *vs_next;		/* next timestamp */
};

/*
 * Vegas specific routines.
 */
int vegastime();

void vegas_init();

void vegas_send();

void vegas_receive();

int vegas_expire();

#endif /*!_netinet_tcp_vegas_h*/
