
/*
 *  $Id: if_hit.c,v 2.4 1995/03/12 01:10:23 lyan Exp lyan $
 */

/*
 * if_hit.c
 *
 *             original skeleton Thomas SKibos 
 *			of UIUC     skibo@cs.uiuc.edu
 *
 *             revamped for WAN emulation by 
 *			Elliot Limin Yan
 *			      of USC	
 *                     lyan@caldera.usc.edu 
 *
 *             RED code is modified from Curtis Villamizar's source. 
 *
 * This is a small device driver that intercepts packets headed for
 * a particular destination and delays them a specified amount of time.
 * It can also randomly drop a fraction of them.  In addition, it also does 
 * provide the illusion of smaller BW.
 */


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioccom.h>
#include <sys/mbuf.h>
#include <sys/errno.h>
#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <sunif/if_hit.h>

static char *sccs_id = "@(#)if_hit.c    ";

/*
 *  This block of statics are the basic parameters of the RED
 *  algorithm (modified a bit).
 */


struct Red 
{
    short current_queue;
    short stale_drop_counter;
    int save_avg;
    int save_sl;
    short was_idle;
    short pkt_counter;
    int last_random;

/*
 *  These parameters are used in the magic Approx[R/pb] used in RED
 */

    short qmax;
    short qmin;
    short qhalf;
    short c1;
    short sh1;
    short c2;
    short sh2;
    unsigned long s;
    
    unsigned long q_size;	/* for sanity check */
};


/* 
 *  	#define ATRACE
 *	#define TIMETEST
 *  	#define TIMING
 *    	#define Q_STATUS
 */

/* hit-box state variables */
#define SIZE 		101		/* size of the address table	*/
#define IF_SIZE 	7		/* size of the interface table  */
#define SEND_Q_SIZE	600
#define TIME_TO_SEND_FUDGE	3
#define NULL 		0

extern int hz;
int gateway_mode = 0;


/* hit-box packet queue */

struct pkt_queue
{
    struct  mbuf    *m;
    struct  in_addr dst;
    n_time  time_to_send;
	
};


struct send_q
{
    n_time       send_time;
    unsigned int packet_size;
};	

/*   interfaces	for which the outgoing packets we intercept */

struct interface
{
    int count;				/* # of connections sharing the 
					 * interface			
					 */

    int	prev_trans_delay;		/* transmission delay of the 
					 *  last packet 	
					 */
    n_time  prev_arrival_time;		/* arrival time of previous 
					 *  packet	
					 */

    unsigned long q_size;		/* for transmission queue in 
					 * bytes 	
					 */
    unsigned long q_len;	
    unsigned int  q_head;
    unsigned int  q_tail;	
  
    struct ifnet *ifhit_ifp;		/* interface we're intercepting 
					 */

    int (*ifhit_real_output)();		/* ditto	
					 */

    int	queue_mgr;
    unsigned int	hi_q_mark;
    
    struct send_q send_queue[SEND_Q_SIZE];	
    struct interface *next;
    struct Red	red;
};

struct interface *if_book[IF_SIZE] = { 0 };


/*    addresses for which packets we intercept	
 */

struct addr_if
{
    struct in_addr ifhit_dest;		/* address we're intercepting	   
					 */
    struct interface *if_ptr;
	
    int	ifhit_pkt_queue_in;
    int	ifhit_pkt_queue_out;

    int	ifhit_p_drop;			/* rate of droping packets, 
					 * multifly 65536 
					 */
    int	ifhit_n_drops;			/* packets dropped since last 
					 * configure 	
					 */
    int	ifhit_rand_drops;	
    int	ifhit_g_drops;			/* FIFO drops	*/
    int	ifhit_red_drops;		/* RED drops	*/
    int	ifhit_delay;			/* delay time to hold packets 	    
					 */

    int	ifhit_ticking;			/* is timeout scheduled? 	    
					 */

    struct pkt_queue ifhit_pkt_queue[ IFHIT_MAX_QUEUE ];
    u_int   ifhit_bandwidth;		/* in bytes / ms 		    
					 */
    struct  addr_if *next;
};

struct addr_if  *addrbook[SIZE] = { 0 };



/*	The following is ported from 
 *	curtis, modified for emulator	- Elliot
 */
/*
 *  EXTERNAL USAGE:
 *
 *  At init time call set_qmax(int x) with the effective total
 *  queueinq capacity for the transmit side on this interface.
 *
 *  When a packet comes in of the bus, call red_ipkt().  If it returns
 *  nonzero keep the packet.  If it is zero discard the packet and
 *  reuse the buffer(s).
 *
 *  After transmit is completed (not queued for transmit), call red_opkt().
 *
 *  If the calls to red_ipkt or red_opkt are not properly balanced,
 *  things won't work well.
 *
 */

/*
 *  set_qmax(int x)
 *
 *  This function should be called at initialization time.  The one
 *  argument should be the value of the effective transmit side queue
 *  capacity for this interface.  This thing sets up all the
 *  parameters for the magic Approx[R/pb] calculation.  How these
 *  numbers come about needs to be explained.
 */

void set_qmax(r, x)
    register struct Red *r;
    register int	x;
{
    int qidx, tmp1, tmp2;
  
    r->qmax = x;
    r->qmin = x >> 5;			/* don't even start counting until 1/32 que */
    for (r->qhalf = (1<<14), qidx = 14; 
	 r->qhalf && !(r->qmax & r->qhalf); r->qhalf >>= 1)
      --qidx;				/* infinite loop if qmax = 0 */
    if (r->qmax == r->qhalf) 
    {
	r->qhalf >>= 1;
	--qidx;
    }

    tmp1 = r->qhalf / (14 - (16 - qidx));
    for (r->sh2 = 16; !(tmp1 & (1 << r->sh2)); --r->sh2)
	;				/* infinite loop if x < 4 */
    --r->sh2;
    r->c2 = 14 - (r->qmax >> r->sh2);
    tmp1 = r->qmax << 2;
    for (r->c1 = 16; !(tmp1 & (1 << r->c1)); --r->c1)
	;
    r->c1 = 16 - r->c1;
    r->sh1 = qidx;
    r->qhalf = (1<<16) * (r->c2 - r->c1) / ((1 << (16 - r->sh1)) - 
					    (1 << (16 - r->sh2)));
}

/*
 *  This is the random number generator.  The variable "s" is the 31
 *  bit random number.  The return value is a scaled integer in the
 *  range of 1/2 to 1, which is what our variation of RED wants.
 */

#define A0 16807		/* magic cookies */
#define S0 40			/* don't change them */

static int quick_red_random(r)
    register struct Red *r;
{

/*    static unsigned long s = S0; */
    register unsigned long p, q;

    q = (r->s & 0x7fff) * A0;
    p = (r->s >> 15) * A0;
    r->s = (p + q) & 0x7fffffff;
    return ((r->s & 0x7fff) | 0x8000);	/* scaled integer in the range 1/2 - 1 */
}

/*
 *  WQ - is a weighting factor described in "Notes on the Holt-Winters
 *  Procedure" ftp://ftp.ee.lbl.gov/papers/holt.ps.Z by Sally Floyd.
 *
 *  WQ2 is WQ/2
 *  WQm1(x) is ((1 - WQ) * x)
 *  WQ2m1(x) is ((1 - WQ2) * x)
 *  WQmult(x) is (WQ * x)
 *  WQ2mult(x) is (WQ2 * 2)
 *
 *  WQ is 1/512 to make the math easy
 */
#define SHIFTY		16	/* use scale integer math scaled this much */
#define WQsh		9	/* shift 9 bits for WQ */
#define WQ2sh		10	/* shift 10 bits for WQ2 */
#define WQm1(x)		((x) - ((x) >> WQsh))
#define WQ2m1(x)	((x) - ((x) >> WQ2sh))
#define WQmult(x)	((x) >> WQsh)
#define WQ2mult(x)	((x) >> WQ2sh)

/*
 *  This thing does the Holt Winters filtering
 */

static int red_pktavg(r)
    register struct Red *r;
{
    register int avg, sl, que;

    /* the following code implements the average queue size filter */
    avg = r->save_avg;
    sl = r->save_sl;
    que = r->current_queue << SHIFTY;
    avg = (WQm1(avg + sl)) + (WQmult(que));
    sl = (WQ2m1(sl)) + (WQ2mult(avg - r->save_avg));
    r->save_avg = avg;
    r->save_sl = sl;

    return avg;
}

/*
 *  red_ipkt()
 *
 *  This function is called when a packet is received on the input
 *  side of the transmitting card -- that is off the MCA bus.  The
 *  return value of the function indicates whether to keep or drop the
 *  packet, a nonzero value indicating keep it, and a zero value
 *  indicating drop it.  There are no arguments to this function and
 *  only the static variables in this module are used.
 */

#define KEEP_IT		1	/* return this to keep the packet */
#define DROP_IT		0	/* return this to drop the packet */

/*
int red_ipkt(r, packet_size)
    register struct Red *r;
    register int packet_size;
*/


int red_ipkt(r)
    register struct Red *r;
{
    int avg, fudge;
    

/*
    r->current_queue+=packet_size;  
    if ( r->current_queue <= packet_size)
*/
    if ( ++r->current_queue <= 1)
    {
	/* queue was empty and in effect still is */
	return KEEP_IT;
    }

    if (r->was_idle) 
    {
	/* note: no way to know how long we were idle */
	r->was_idle = 0;
	r->save_avg >>= 1;
	r->save_sl = 0;
	r->pkt_counter = -1;
	if (r->qmax < 0)
	    set_qmax(r, r->q_size>>9);
    }

    avg = red_pktavg(r);
    /* adjust the result down from the scale integer math */
    avg >>= SHIFTY;
    /* minimum drop threshhold */
    if (avg <= r->qmin) 
    {
	/* this one gets by unharmed - and the counter is reset */
	r->pkt_counter = -1;
	return KEEP_IT;
    }
    /* determine whether we need a new count */
    ++r->pkt_counter;
    if (r->pkt_counter <= 0) 
    {
	if (r->stale_drop_counter)
	    r->last_random = quick_red_random(r);
	r->stale_drop_counter = 0;
	return KEEP_IT;
    }
    /* this is the magic Approx[R/pb] in RED */
    if (avg >= r->qmax) 
    {
	fudge = 14;
    } 
    else if (avg > r->qhalf) 
    {
	fudge = r->c2 + (avg >> r->sh2);
    } 
    else 
    {
	fudge = r->c1 + (avg >> r->sh1);
    }
    if (r->pkt_counter < (r->last_random >> fudge)) 
    {
	/* this one gets by unharmed */
	return KEEP_IT;
    }
    /* NUKE IT! */
    r->pkt_counter = -1;
    r->stale_drop_counter = 1;

/*
    r->current_queue-=packet_size;

*/
    --r->current_queue;
    return DROP_IT;
}

/*
void red_opkt(r, packet_size)
    register struct Red *r;
    register int packet_size;
*/

void red_opkt(r)
    register struct Red *r;
{
/*
    if ((r->current_queue-=packet_size) == 0) 
*/
    if (--r->current_queue == 0) 
    {
	/* queue just emptied completely */
	r->was_idle = 1;
	return;
    }
    red_pktavg(r);
}


void print_qmax(r)
    register struct Red *r;
{
    printf
	("max= %d, half=%d, s1= %d, s2= %d, q(0)= %d, q(1/2)= %d %d, q(max)= %d\n",
	 r->qmax, r->qhalf, r->sh1, r->sh2,
	 1<<(16-r->c1),
	 1<<(16-(r->c1+(r->qhalf>>r->sh1))),
	 1<<(16-(r->c2+(r->qhalf>>r->sh2))),
	 1<<(16-(r->c2+(r->qmax>>r->sh2))));
}

print_qfunct(r)
    register struct Red *r;
{
    int i, v;
    print_qmax(r);
    for (i = 0; i <= r->qmax; ++i) 
    {
	if (i > r->qhalf)
	    v = 1 << (16 - (r->c2 + (i >> r->sh2)));
	else
	    v = 1 << (16 - (r->c1 + (i >> r->sh1)));
	if (!i)
	    ;
	else if (!(i & 15))
	    printf("\n");
	else
	    printf(" ");
	printf("%4d", v);
    }
    printf("\n");
}

/*	end of RED code
 */


/* hash routines for the table of interfaces	
 */

unsigned if_hash(ifp)			
    register struct ifnet *ifp;
{
    return ((unsigned)ifp)%IF_SIZE;
}


/* hash lookup for interface table 
 */

struct interface *if_lookup(ifp)
    register struct ifnet *ifp;
{
    int     i;
 
    struct interface *np;

    for (np=if_book[if_hash(ifp)]; np != NULL; np=np->next)
	if ( ifp == np->ifhit_ifp)
	    return np;
    return NULL;
}


/* hash install for interface table
 * hash by the interface ifnet
 * store the real if_output routine
 * increment the count if the ifnet is previously hijacked
 */

struct interface *if_install(ifp, if_output)
    register struct ifnet *ifp;
    register int (*if_output)();
{
    struct interface *np;
    unsigned hashval;
    
    if ((np = if_lookup(ifp)) == NULL)
    {
	if ((np = (struct interface *) kmem_alloc(sizeof(*np)))== NULL)
	    return NULL;
	
	bzero(np, sizeof(*np));
	hashval = if_hash(ifp);
	np->next = if_book[hashval];
	if_book[hashval] = np;
    
	np->ifhit_ifp = ifp;
	np->ifhit_real_output = if_output;
	np->count++;
	
	np->prev_trans_delay = 0;
	np->prev_arrival_time = 0;
	np->q_head = 0;
	np->q_tail = 0;
	np->q_len  = 0;
	
/* RED stuff */	
	np->red.current_queue = 0;
	np->red.stale_drop_counter = 1;
	np->red.save_avg = 0;
	np->red.save_sl = 0;
	np->red.was_idle = 1;
	np->red.pkt_counter = -1;
	np->red.last_random = 0;
	np->red.qmax = -1;
	np->red.qmin = 0;
	np->red.qhalf = 0;
	np->red.c1 = 0;
	np->red.sh1 = 0;
	np->red.c2 = 0;
	np->red.sh2 = 0;
	np->red.s = S0;
	
    }
    else
	np->count++;

    return np;
}


/* delete the interface from the table
 * if more than one connection exists
 * decrement the count, else
 *  restore the original if_output
 */

if_delete( if_ptr )
    register struct interface *if_ptr;
{
    struct interface *np, *pnp;

    pnp = NULL;
  
    for (np = if_book[if_hash(if_ptr->ifhit_ifp)]; 
	 np != NULL; np = np->next)
    {
	if ( np->ifhit_ifp == if_ptr->ifhit_ifp )
	    break;
	pnp = np;			/* separate chaining stinks */
    }

    if (np != NULL)
    {
	if (np->count > 1)
	    np->count--;
	else
	{
	    if (pnp != NULL)
		pnp->next = np->next;
	    else
		if_book[if_hash(if_ptr->ifhit_ifp)] = np->next;
      
	    if_ptr->ifhit_ifp->if_output = if_ptr->ifhit_real_output;
	    if_ptr->ifhit_ifp = (struct ifnet *)0;	    
	    kmem_free((caddr_t)np, sizeof(*np));
	}
    }
}

/* hash routines for table of addresses 
 * address book is hashed by the address of the destination 
 */

unsigned addr_hash( dest )
    register struct in_addr dest;
{
    return ((unsigned)dest.s_addr)%SIZE;
}


/* address lookup */

struct addr_if *addr_lookup( dest )
    register struct in_addr dest;
{
    struct addr_if  *np;

    for (np = addrbook[addr_hash(dest)]; np != NULL; np = np->next)
	if (dest.s_addr == np->ifhit_dest.s_addr)
	    return np;
    return NULL;
}


/* address install 
 * if address is previously intercepted
 * change the parameter
 * else add a new entry to the address table and interface table
 */


struct addr_if *addr_install( req )
    register struct if_hitreq *req;
{	
  
    struct addr_if *np;
    unsigned hashval;
    struct route ro;
    int     s;
    struct ifnet *ifhit_ifp = (struct ifnet *)0; /* interface we're 
						  * intercepting 
						  */
    int	ifhit_output();
    int	(*ifhit_real_output)();
  
    struct ifnet *if_real;
    
    s = splimp();
		
    if ((np = addr_lookup(req->dest)) == NULL)
    {
	if ((np = (struct addr_if *) kmem_alloc(sizeof (*np))) == NULL)
	    return NULL;
    
	bzero(np, sizeof(*np));
	hashval = addr_hash(req->dest);
	np->next = addrbook[hashval];
	addrbook[hashval] = np;
	
    
	/* hijack the network interface 
	 */

    {
	register struct sockaddr_in *dst=	
	    (struct sockaddr_in *) &ro.ro_dst;
	
	dst->sin_family = AF_INET;
	dst->sin_addr = req->dest;
	
    }
		
	/* FIND route */

	ro.ro_rt = (struct rtentry *) 0;
	rtalloc( &ro );
		
	if ( ro.ro_rt == 0 )
	{
	    splx(s);
	    return NULL;
	}

	ifhit_ifp = ro.ro_rt->rt_ifp;
		
	ifhit_real_output = ifhit_ifp->if_output;		
	ifhit_ifp->if_output = ifhit_output;
	np->ifhit_dest.s_addr = req->dest.s_addr;
	np->ifhit_ticking = 0;

	/* add to the interface table 
	 */

	np->if_ptr = if_install(ifhit_ifp, ifhit_real_output); 
	
	RTFREE( ro.ro_rt );
		
	/* Initialize queue
	 */
	np->ifhit_pkt_queue_in = 0;
	np->ifhit_pkt_queue_out = 0;

    }
    
    np->ifhit_p_drop = req->p_drop;
    np->ifhit_delay = (req->delay*hz)/1000;

    np->ifhit_bandwidth =  req->bandwidth;
    np->ifhit_rand_drops = 0;
    np->ifhit_n_drops = 0;
    np->ifhit_g_drops = 0;
    np->ifhit_red_drops = 0;


    np->if_ptr->q_size = req->qsize;
    np->if_ptr->hi_q_mark = (np->if_ptr->q_size >> 1) + 
	(np->if_ptr->q_size >> 2);
    np->if_ptr->queue_mgr = req->queue_mgr;
    np->if_ptr->red.q_size = req->qsize;	   

/* init RED queue */
    if (req->queue_mgr == RED)
    {

	set_qmax(&np->if_ptr->red, np->if_ptr->q_size>>9);
/*
	set_qmax(&np->if_ptr->red, np->if_ptr->q_size>>9);
	print_qfunct(&np->if_ptr->red);
*/
    }

    splx(s);
    return np;
}


/* delete an address from the address book 
 */

addr_delete( dest )
    register struct in_addr dest;
{
    struct addr_if *np, *pnp;
    int s;

    pnp = NULL;

    for (np = addrbook[addr_hash(dest)]; np != NULL; np = np->next)
    {
	if (dest.s_addr == np->ifhit_dest.s_addr)
	    break;
	pnp = np;
    }

    if (np !=NULL)
    {
	if (pnp != NULL)
	    pnp->next = np->next;
	else
	    addrbook[addr_hash(dest)] = np->next;
	
	/* will delete the interface if no other connection is 
	 * sharing the 
	 * interface as well
	 */

	if_delete(np->if_ptr);		
	kmem_free((caddr_t)np, sizeof(*np));
    }
}


/*
 * handle ioctl() requests to hit-box device.
 */

ifhit_ioctl( dev, cmd, ifreq, flags )
    dev_t   dev;
    int     cmd;
    union {
	struct if_hitreq hitreq;
	struct if_hitstats stats;
	int	gateway_mode;
    } *ifreq;
    int     flags;
{
    int     error = 0, s;
    struct addr_if *np;
	
    switch (cmd) 
    {
    case HITIOCTL_OFF:
	/* Restore old if_output routine.
	 */
	s = splimp();
	restore_all_interfaces();
	splx(s);
	break;
	
    case HITIOCTL_ON:
	/* Configure a new destination to "hit."
	 */

	/* Zero parameters?  restore interface.
	 */

	if (ifreq->hitreq.p_drop == 0 && 
	    ifreq->hitreq.delay == 0 && 
	    ifreq->hitreq.bandwidth == 0)
	{
	    s = splimp();
      
	    restore_interface(ifreq->hitreq.dest);
	    splx(s);
	    break;
	}

	/*  it is impossible to have 0 for bandwidth
	 *  0 delay for network
	 *  Nothing is changed
	 */

	if (ifreq->hitreq.bandwidth == 0 || 
	    ifreq->hitreq.delay == 0 )
	    break;

	if (addr_install(&ifreq->hitreq) == NULL)
	    return EHOSTUNREACH;
	
	break;

    case HITIOCTL_STATS:
	if ( (np = addr_lookup(ifreq->stats.dest)) != NULL ) 
	{
	    ifreq->stats.p_drop = np->ifhit_p_drop;
	    ifreq->stats.n_drops = np->ifhit_n_drops;
	    ifreq->stats.rand_drops = np->ifhit_rand_drops;
	    ifreq->stats.g_drops = np->ifhit_g_drops;
	    ifreq->stats.red_drops = np->ifhit_red_drops;
	    ifreq->stats.delay = (np->ifhit_delay*1000)/hz;
	    ifreq->stats.bandwidth = np->ifhit_bandwidth;
	    ifreq->stats.if_used = (int)np->if_ptr->ifhit_ifp;
	    ifreq->stats.connection_count = np->if_ptr->count;
	    ifreq->stats.qsize= np->if_ptr->q_size;
	    ifreq->stats.qlen = np->if_ptr->q_len;
	    ifreq->stats.q_mgr = np->if_ptr->queue_mgr;
	    ifreq->stats.red_c_queue = np->if_ptr->red.current_queue;
	    ifreq->stats.red_save_avg = np->if_ptr->red.save_avg;
	    ifreq->stats.red_q_max = np->if_ptr->red.qmax;
	}
	else 
	    bzero( &ifreq->stats,sizeof(struct if_hitstats));
	break;

    case HITIOCTL_MODE:
	gateway_mode = ifreq->gateway_mode; 
	break;

    default:
	error = EINVAL;
	break;
    }

    return error;
}	



/* restore interface for an address 
 */

static restore_interface( dest )
    register struct in_addr dest;
{
    int ifhit_timeout();

    struct addr_if *np;

    if ((np = addr_lookup(dest)) != NULL)
    {

	/* kill the timeout 
	 */

	untimeout( ifhit_timeout, (caddr_t)np );

	np->ifhit_ticking = 0;
                
	/* Free up any queued packets.
	 */
		
	while ( np->ifhit_pkt_queue_in != np->ifhit_pkt_queue_out )
	{
	    m_freem(np->ifhit_pkt_queue[ np->ifhit_pkt_queue_out ].m );
	    np->ifhit_pkt_queue_out =
		(np->ifhit_pkt_queue_out+1) % IFHIT_MAX_QUEUE;
	}

	
	/* delete address from the address book 
	 */

	addr_delete( dest );
		
    }
}

restore_all_interfaces()
{	
    int i;
	
    gateway_mode = 0;

    for (i = 0; i < SIZE; i++)
    {
	while (addrbook[i] != NULL)	/* this is beautiful */
	    restore_interface(addrbook[i]->ifhit_dest);
    }	
}


round_div(num, denom)
    register int num;
    register int denom;
{
    register int a, b;

    a = num/denom;
    b = num%denom;

    if ( (b <<1) >= denom)
	return a+1;
	
    return a;
}



/* The output interface the intercepts packets that goes to
 * into the network interface.  It puts the packet on a
 * queue and delays its send time according to the delay,
 * bandwidth specification.  Queueing delay is considered in
 * the computation of delay.
 */

ifhit_output( ifp, m, dst )
    register struct ifnet    *ifp;
    register struct mbuf     *m;
    register struct sockaddr *dst;
{
    register struct ip 	*ipp;
    static u_int random;
    int ifhit_timeout();

    register struct addr_if *np;
    register struct interface *if_ptr;
    register int real_delay;
    register int trans_delay;		/* transmission delay 	*/
    register int queue_delay;		/* queuing delay	*/
    register unsigned int tmp_q_size;
    register n_time arrival_time;
    register struct sockaddr_in	*gw;

    if ( random == 0 )
	random = (u_int) iptime();
    random = (random & 0x80000000) ? random<<1 ^ 0x54a5 :  random<<1;

    
    /* Use address of FINAL destination, not dst (which is gateway
     * to route packet).
     */
	
    ipp = mtod( m, struct ip * );
	
    /*
     *	what in the bloody world is gateway_mode ?
     *  as long the link for packet's next hop is set for certain 
     *  delay, BW, drop, etc., gateway_mode is that ALL packets 
     *  with this next hop are delayed according to the above parameters set.
     *  In NON gateway_mode, only end-2-end delay is considered.
     *  Thus in gateway mode, all emulation parameters are
     *  additive. 
     *
     *			A
     *			 \
     *		    20 ms \   10ms
     *                      B ----- C
     *	
     *	Assuming packets are coming from A -> C, passing B, where delay is 
     *  marked as above, in gateway_mode, all packets from A -> C will have
     *  30 ms of delay.  In NON gateway_mode, the delay is set only between
     *  A ans C, independent of B and their interconnecting link speed.  Almost 
     *  always, we use gateway_mode for emulator experiments.  In non-gateway mode
     *  if we set A -> B be 20 ms, all packets from A->B will be delayed 20 ms
     *  and packets A->C will NOT experience this 20ms, but in gateway mode
     *  all packets will from A->C will experience the 20 ms in addition to 
     *  packets from A->B.
     */

    if (gateway_mode)	       
    {	
	gw = (struct sockaddr_in *)dst; 
	np = addr_lookup(gw->sin_addr);	
    }
    else
	np = addr_lookup(ipp->ip_dst);
	

    /* if the address is hijacked 
     */
    
    if (np != NULL  && dst->sa_family == AF_INET)
    {
	if ((random&65535) < np->ifhit_p_drop ) 
	{
	    /* Drop this packet.
	     */

	    m_freem( m );
	    np->ifhit_rand_drops++;
	    return 0;
	}
	else if (np->ifhit_delay == 0 && 
		 np->ifhit_bandwidth == 0)
	    /* Forward it to its real interface, now.
	     */
	    return (*(np->if_ptr->ifhit_real_output))( ifp, m, dst );
	else 
	{
	    /* Queue packet for delay.
	     * Will packet fit?  If not, drop it.
	     */

	    if ((np->ifhit_pkt_queue_in+1)%IFHIT_MAX_QUEUE ==
		np->ifhit_pkt_queue_out ) 
	    {		
		/* Drop this packet.
		 */
		m_freem( m );
		np->ifhit_n_drops++;
		return 0;
	    }

	    arrival_time = ntohl(iptime());

	    /* Any queueing disiplines goes here
	     * FIFO Stuff 		 
	     * RED STUFF
	     */

	    if (np->if_ptr->q_size != 0) 	/* if router queue size is not infinite */
	    {		
		if (np->if_ptr->queue_mgr == RED)        /* RED queue */
		{
		    /* clean up router queue */
		    while (np->if_ptr->q_head != np->if_ptr->q_tail)
		    {
			if (np->if_ptr->send_queue[np->if_ptr->q_head].send_time <= 
			    arrival_time )
			{		
			    np->if_ptr->q_len -= 
				np->if_ptr->send_queue[np->if_ptr->q_head].packet_size;
			    red_opkt(&np->if_ptr->red);
				 
			    np->if_ptr->q_head = 
				(np->if_ptr->q_head+1) % SEND_Q_SIZE;
			}	
			else
			    break;
		    }
				
		    if (red_ipkt(&np->if_ptr->red) == DROP_IT) /* Random early Drop */
		    {
			m_freem( m );
			np->ifhit_red_drops++;
			return 0;
		    }

		    if ((tmp_q_size = (np->if_ptr->q_len + ntohs(ipp->ip_len))) > 
			np->if_ptr->q_size )		/* if router queue is full */
		    {
			m_freem( m );
			np->ifhit_g_drops++;
			return 0;
		    }

		    /* append packet to the router queue */
		    np->if_ptr->q_len = tmp_q_size;
		    np->if_ptr->send_queue[np->if_ptr->q_tail].packet_size = 
			ntohs(ipp->ip_len);
		}
		else		/* FIFO queue */
		{	
    		    /* clean up router queue */
		    if ((tmp_q_size = (np->if_ptr->q_len + ntohs(ipp->ip_len))) > 
			np->if_ptr->hi_q_mark)	/* periodically clean the queue */
		    {
			while (np->if_ptr->q_head != np->if_ptr->q_tail)
			    if (np->if_ptr->send_queue[np->if_ptr->q_head].send_time
				<= arrival_time )
			    {	
				np->if_ptr->q_len -= 
				    np->if_ptr->send_queue[np->if_ptr->q_head].packet_size;
				np->if_ptr->q_head = 
				    (np->if_ptr->q_head+1) % SEND_Q_SIZE;
			    }
			    else
				break;
					
			if ((tmp_q_size = 
			     (np->if_ptr->q_len + ntohs(ipp->ip_len))) > 
			    np->if_ptr->q_size )	/* if router queue is full */
			{
			    m_freem( m );
			    np->ifhit_g_drops++;
			    return 0;
			}
		    }
			
		    /* append packet to the router queue */
		    np->if_ptr->q_len = tmp_q_size;
		    np->if_ptr->send_queue[np->if_ptr->q_tail].packet_size = 
			ntohs(ipp->ip_len);
		}		
	    }

	    /* Queue the packet to the outgoing pkt_queue, note different
	     * from router queue above
	     */
	    
	    np->ifhit_pkt_queue[ np->ifhit_pkt_queue_in ].m = m;
	    np->ifhit_pkt_queue[ np->ifhit_pkt_queue_in ].dst.s_addr =
		(( struct sockaddr_in *) dst)->sin_addr.s_addr;

	    if (arrival_time < np->if_ptr->prev_arrival_time)	
	    {
		np->if_ptr->prev_arrival_time = 0;
		np->if_ptr->prev_trans_delay = 0;
		queue_delay = 0;
	    }
	    else
	    {
		if ((queue_delay = np->if_ptr->prev_trans_delay 
		     - (arrival_time - np->if_ptr->prev_arrival_time)) < 0)
		    queue_delay = 0;
	    }

	    /* the packet's transmission delay is 
	     * queueing_delay + actual_xmt_delay
	     */

	    trans_delay = 
		queue_delay +  round_div(ntohs(ipp->ip_len),np->ifhit_bandwidth);
	    
	    if (np->if_ptr->q_size != 0)	
	    {
		np->if_ptr->send_queue[np->if_ptr->q_tail].send_time = 
		    arrival_time + trans_delay;
		np->if_ptr->q_tail = (np->if_ptr->q_tail+1)% SEND_Q_SIZE;		
	    }
	    
/*
	    real_delay = 
		np->ifhit_delay + trans_delay - 
		    round_div(ntohs(ipp->ip_len),1250);
*/
	    
	    real_delay = np->ifhit_delay + trans_delay;
			   
	    np->ifhit_pkt_queue[ np->ifhit_pkt_queue_in ].time_to_send =
		arrival_time + real_delay;

	    np->ifhit_pkt_queue_in = 
		(np->ifhit_pkt_queue_in+1)%IFHIT_MAX_QUEUE;
	   
	    if ( !np->ifhit_ticking ) 	/* if timeout not installed */
	    {
		/* It should be 
         	 * timeout(ifhit_timeout, (caddr_t)np, real_delay*hz/1000);
		 * but since the machine is set at hz = 1000 (instead of 100) ...
		 */

		timeout(ifhit_timeout, (caddr_t)np, real_delay );
		np->ifhit_ticking = real_delay;
	    }

	    np->if_ptr->prev_arrival_time = arrival_time;
	    np->if_ptr->prev_trans_delay  = trans_delay;
			
	    return 0;
	    
	}
	
    }
    else
	/* Forward it to its real interface.
	 */
    {
	/* if the interface is hijacked but not a internet packet */
	
	if ((if_ptr = if_lookup(ifp)) != NULL)	
	    return (*if_ptr->ifhit_real_output)(ifp, m, dst);
	else				/* go to an unhijacked interface */
	    return (*ifp->if_output )( ifp, m, dst );

    }
         
}


/* timeout routine 
 * input param is the intercepted address interface
 * 
 * the interrupt handler that forwards packets to real
 * net interface, and schedules the remaining packets 
 * in the pkt_queue.
 */

ifhit_timeout( param )
    register caddr_t	param;
{
    int s;
    register struct sockaddr_in dst;
    register struct addr_if *np = (struct addr_if *)param;
    register int real_delay;
    register struct ip *ipp;
    register n_time timer;
    
    dst.sin_family = AF_INET;
 
    s = splnet();

    timer = ntohl(iptime()) + TIME_TO_SEND_FUDGE; 
 
    /* Forward packets to real interface */

 keepsending:				
    while (np->ifhit_pkt_queue_out != np->ifhit_pkt_queue_in &&	       
	   np->ifhit_pkt_queue[ np->ifhit_pkt_queue_out ].time_to_send <= timer ) 
    {		
	dst.sin_addr.s_addr = 
	    np->ifhit_pkt_queue[ np->ifhit_pkt_queue_out ].dst.s_addr;
    
	/* Send packets */

	(*np->if_ptr->ifhit_real_output)
	    (np->if_ptr->ifhit_ifp,
	     np->ifhit_pkt_queue[ np->ifhit_pkt_queue_out ].m, &dst );
    
	np->ifhit_pkt_queue_out = (np->ifhit_pkt_queue_out+1)%IFHIT_MAX_QUEUE;
    }

    /* Schedule next packet in the pkt_queue */

    if ( np->ifhit_pkt_queue_out != np->ifhit_pkt_queue_in ) 
    {
	ipp = mtod(np->ifhit_pkt_queue[ np->ifhit_pkt_queue_out ].m, 
		   struct ip * );	

	timer = ntohl(iptime());
	real_delay = 
	    np->ifhit_pkt_queue[ np->ifhit_pkt_queue_out].time_to_send - timer;
	
	if (real_delay <= 0)
	    goto keepsending;

	/* Else, schedule the pkt*/

	/* It should be 
	 * timeout(ifhit_timeout, param, real_delay*hz/1000);
	 * but since the machine is set at hz = 1000 (instead of 100) ...
	 */
	
	timeout( ifhit_timeout, param, real_delay );
	np->ifhit_ticking = real_delay;
    }
    else
	np->ifhit_ticking = 0;

    splx(s);
}


