/*
 * if_hit.h
 *
 * Header file for network hit-box.
 *
 * @(#)if_hit.h	 
 */

/* structure used for requesting the hit-box.
 */
struct if_hitreq 
{
    struct in_addr 	dest;		/* destination to hit, or INADDR_ANY */
    int			p_drop;		/* probability of a drop times 65536 */
    int			delay;		/* delay on packets sent, in ms.  */
    u_int		bandwidth;
    unsigned long	qsize;
    int			queue_mgr;
};

/* structure used for returning stats on the hit-box.
 */
struct if_hitstats 
{
    struct in_addr dest;		/* current destination */

    int		p_drop;			/* current drop rate times 65536 */
    int		delay;			/* average time packets held, in ms */
    int 	bandwidth;
	
    int		n_drops;		/* number of packets dropped since
					 *  last HITIOCTL_ON call. natural
					 *  drops
					 */
    int		rand_drops;
    int		g_drops;		/* FiFo Gateway drops	
					 */
    int		red_drops;		/* RED drops
					 */

    int  	if_used;		/* if_addr		*/
    int		connection_count;	/* number of connection sharing the if	*/
    int		gateway_mode;

    unsigned long	qsize;
    unsigned long	qlen;
    
    short	red_c_queue;
    int		red_save_avg;
    int		red_q_max;
    int		q_mgr;
};

int	gateway_mode;

/* For ioctl() requests:
 */
#define HITIOCTL_OFF	_IO(h, 0)			/* Turn off hitbox */
#define HITIOCTL_ON	_IOW(h, 1, struct if_hitreq)	/* Turn on hitbox & */
/* specify params */
#define HITIOCTL_STATS	_IOWR(h, 2, struct if_hitstats)	/* get hit-box stats */
#define HITIOCTL_MODE	_IOW(h, 3, int)

/* Constants
 */
#define IFHIT_MAX_QUEUE	2000
#define P_DROP_FACTOR	65536

#define INF_FIFO -1
#define FIFO	0
#define RED	1
#define FQ	2
#define WFQ	3
#define VC	4
#define	SVC	5
